// VinciaAntenna.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2014 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the VinciaAntenna header file
#include "VinciaAntenna.h"

using namespace Pythia8;

namespace Vincia {
  
  //**********************************************************************

  //-----------
  // Antenna
  // Base class definitions not found in header
  
  bool Antenna::init(bool isSectorIn, int minMaxVar) {
      
    // Check whether pointers are initialized
    if (!isInitPtr) return false;

    // Verbose level
    verbose              = settingsPtr->mode("Vincia:verbose");
    
    // Antenna on/off
    isOnSav              = settingsPtr->flag(vinciaName());
    
    // Antenna type
    isSectorSav          = isSectorIn;

    // Terms on/off
    useFiniteTermsSav    = settingsPtr->flag("Vincia:useFiniteTerms");
    useCollinearTermsSav = settingsPtr->flag("Vincia:useCollinearTerms");

    // Charge factor
    chargeFactorSav      = settingsPtr->parm(vinciaName()+":chargeFactor");

    // Kinematics map type (first look for specific, else default to global)
    if (settingsPtr->isMode(vinciaName()+":kineMapType"))
      kineMapTypeSav     = settingsPtr->mode(vinciaName()+":kineMapType");
    else
      kineMapTypeSav     = settingsPtr->mode("Vincia:kineMapType");

    // Initialize user-defined finite terms (minMaxVar = 0)
    Cs00 = settingsPtr->parm(vinciaName()+":Cs00");
    Cs10 = settingsPtr->parm(vinciaName()+":Cs10");
    Cs01 = settingsPtr->parm(vinciaName()+":Cs01");
    Cf00 = settingsPtr->parm(vinciaName()+":Cf00");
    Cf10 = settingsPtr->parm(vinciaName()+":Cf10");
    Cf01 = settingsPtr->parm(vinciaName()+":Cf01");
    if (settingsPtr->isParm(vinciaName()+":Cp00"))
      Cp00 = settingsPtr->parm(vinciaName()+":Cp00");
    if (settingsPtr->isParm(vinciaName()+":Cp10"))
      Cp10 = settingsPtr->parm(vinciaName()+":Cp10");
    if (settingsPtr->isParm(vinciaName()+":Cp01"))
      Cp01 = settingsPtr->parm(vinciaName()+":Cp01");

    // Base class max variation is all C00 = 5
    // (Note: should be good enough for most inheriting classes)
    if (minMaxVar >= 1) {
      Cs00 = 5;
      Cf00 = 5;
      Cp00 = 5;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }
    
    // Base class min variation is all Cij = 0 
    // (Note: may give problems with positivity in some cases. May need
    // to be redefined by inheriting classes.)
    else if (minMaxVar <= -1) {
      Cs00 = 0;
      Cf00 = 0;
      Cp00 = 0;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
		      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Collinear partitioning: different for global and sector
    if (!isSectorSav) {
      // Global antennae: alpha is adjustable parameter.
      alphaSav = settingsPtr->parm("Vincia:octetPartitioning");    
    } 
    else {
      // Sector antennae: force alpha = 1 and check sector terms 
      alphaSav = 1.0;
      useSectorTermsSav  =  settingsPtr->flag("Vincia:useSectorTerms");
    }
    
    // Return OK
    isInit = true;
    return isInit;
    
  }
  
  
  // The (dimensionless) antenna function. Base class returns zero.
  double Antenna::antennaFunction(double, double, 
				     double, double, double,
				     int, int, int, int, int) {
    printOut("Antenna::antennaFunction","Error. Dummy function in base "
	     "class called.");
    return 0.0;
  }

  // The (dimensionless) AP kernel. Base class returns -1.
  double Antenna::AltarelliParisi(double, double, double, double, double, 
                                  int, int, int, int, int) {
    return -1.0;
  }

  // Generic check of antenna function during initialization
  bool Antenna::check() {

    bool isOK = true;
    
    // Check soft singularity against massless dipole factor
    if (idJ() == 21 || idJ() == 22) {
      // Test one symmetric and one asymmetric soft phase-space point
      for (int iTest = 0; iTest <= 1; ++iTest) {
        // Test invariants
        double yij = 1.0e-3 * pow(10.0,iTest);
        double yjk = 1.0e-3 / pow(10.0,iTest);
        // Compute eikonal
        double eik = 2.*(1.-yij-yjk)/yij/yjk;
        // Compute antenna
        double ant = antennaFunction(yij,yjk);
        // Compare
        double ratio = ant/eik;
        if (abs(ratio-1) >= 0.001) {
          isOK = false;
          if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                     "(eikonal "+num2str(iTest,1)+")");
        } else if (verbose >= 6) {
          printOut(vinciaName()+":check","OK (eikonal "+
                   num2str(iTest,1)+")");
        }
      }
    }
      
    // Test all helicity configurations for collinearity and posivity
    int hA(9), hB(9), hI(9), hJ(9), hK(9);
    string helString = " 9 9 -> 9 9 9";
    for (int iHel = -1; iHel < 16; ++iHel) {
      if (iHel >= 0) {
        hJ = 2*(iHel%2) - 1;
        hI = 2*((iHel/2)%2) - 1;
        hK = 2*((iHel/4)%2) - 1;
        hA = 2*((iHel/8)%2) - 1;
        hB = 2*((iHel/16)%2) - 1;
        helString = num2str(hA,2) + num2str(hB,2) + " ->"
          + num2str(hI,2) + num2str(hJ,2) + num2str(hK,2);
      }

      // Check collinear singularities against massless AP kernels, if 
      // implemented (base class dummy AP function returns -1)
      if (useCollinearTermsSav) {
        for (int iTest=0; iTest<=2; ++iTest) {
          // Test invariants, for a few points along collinear axis.
          double y1   = 1.0e-5;
          double y2   = 0.2 + iTest*0.3;
          // Compute AP kernels
          // V->FF splitting antennae only have JK collinear singularity
          double AP1  = AltarelliParisi(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
          double AP2  = AltarelliParisi(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);
          if ( idJ() != 21 && idJ() != 22 ) AP1 = -1.;
          // There is only an IJ singularity if hB is conserved 
          if ( hB != hK ) AP1 = -1.;
          if ( hA != hI ) AP2 = -1.;
          // Sector antennae
          if ( isSectorSav ) {
            // Gluon emission: allow -X -> +-X but not -X -> ++X
            if ( idJ() == 21 || idJ() == 22 ) {
              int aCheck = hA*hI + hA*hJ;
              int bCheck = hB*hJ + hB*hK;
              if (aCheck <= -2 || bCheck <= -2) {
                AP1 = -1.;
                AP2 = -1.;
              }
            }
          }
          // Global antennae
          else {
            // Gluon emission: do not allow external helicity flips
            if ( idJ() == 21 || idJ() == 22 ) {
              if ( hA != hI || hB != hK ) {
                AP1 = -1.;
                AP2 = -1.;
              }
            }
          }
          // Are there any limits to check?
          if (AP1 < 0. && AP2 < 0.) continue;

          // Compute antennae
          double ant1 = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
          double ant2 = antennaFunction(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);
          // For global antennae, add "swapped" terms to represent neighbor
          if (!isSectorSav) {
            if (idA() == 21) {
              ant1 += antennaFunction(y1,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
              ant2 += antennaFunction(y2,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
            }
            if (idB() == 21) {
              ant1 += antennaFunction(1.-y1-y2,y2,0.,0.,0.,hA,hB,hI,hK,hJ);
              ant2 += antennaFunction(1.-y1-y2,y1,0.,0.,0.,hA,hB,hI,hK,hJ);
            }
          }
          
          // Compare
          if (AP1 > 0.) {
            double ratio = ant1/AP1;
            if (abs(ratio-1) >= 0.01) {
              isOK = false;
              if (verbose >= 1) 
                printOut(vinciaName()+":check","WARNING (collinear ij "
                         +num2str(iTest,1)+" "+helString+" )");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                    <<num2str(zi(y1,y2),9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear ij "
                       +num2str(iTest,1)+" "+helString+" )");
            }
          }
          if (AP2 > 0.) {
            double ratio = ant2/AP2;
            if (abs(ratio-1) >= 0.01) {
              isOK = false;
              if (verbose >= 1) 
                printOut(vinciaName()+":check","WARNING: FAILED (collinear jk "
                         +num2str(iTest,1)+" "+helString+" )");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant2,9)<<" y1 = "<<num2str(y2,9)
                    <<" y2 = "<<num2str(y1,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                    <<num2str(zk(y1,y2),9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear jk "
                       +num2str(iTest,1)+" "+helString+" )");
            }
          }
        }
      }
      
      // Check positivity
      int nPointsCheck = settingsPtr->mode("Vincia:nPointsCheck");
      bool isPositive  = true;
      bool isZero      = true;
      bool hasDeadZone = false;
      double deadZoneAvoidance = settingsPtr->parm("Vincia:deadZoneAvoidance");
      for (int iTest = 0; iTest < nPointsCheck; ++iTest) {

        // Generate a random point in phase-space triangle
        double y1 = rndmPtr->flat();
        double y2 = rndmPtr->flat();
        if (y1 + y2 > 1.0) {
          y1 = 1 - y1;
          y2 = 1 - y2;
        }
        double ant = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);

        // Check positivity (strict)
        if (ant < 0.) {
          isPositive = false;
          if (verbose >= 3) 
            printOut(vinciaName()+":check","ERROR ant("
                     +num2str(y1,9)+","+num2str(y2,9)+" ; "+helString+") < 0");
        } else if (ant > 0.) {
          isZero = false;
        }
        // Check for dead zones away from phase-space boundary
        if (1-y1-y2 > 0.05 && y1 > 0.05 && y2 > 0.05
            && ant < deadZoneAvoidance) hasDeadZone = true;
        
      } // End nPointsCheck sum for positivity checks

      // OK?
      isOK = isOK && isPositive;

      // Verbose output
      if (!isPositive && verbose >= 1) {
        printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                 +helString+" )");
      } else if (isPositive && !isZero && verbose >= 6) {
        printOut(vinciaName()+":check","OK (is positive "
                 +helString+" )");        
      }
      if (!isZero) {
        if (hasDeadZone && verbose >= 1) {
          printOut(vinciaName()+":check","WARNING (dead zone encountered "
                   +helString+" )");
        } else if (!hasDeadZone && verbose >= 6) {
          printOut(vinciaName()+":check","OK (no dead zones "
                   +helString+" )"); 
        }
      }
      
    } // End helicity sum


    return isOK;
    
  }
  
  //======================================================================
  // Begin inherited classes: one for each type of antenna 

  //------------
  // QQEmit::init()

  bool QQEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = Antenna::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Min Variation
    if (minMaxVar == -1) {
      Cs00 = -4;
      Cs01 = 0;
      Cs10 = 0;
      Cf00 = 2;
      Cf01 = 0;
      Cf10 = 0;
      Cp00 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }
	  
    // Check whether finite terms are symmetric
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cs10 = Cs01 by averaging.");
      Cs10 = 0.5 * (Cs10 + Cs01);
      Cs01 = Cs10;
    }
    if (Cf10 != Cf01) {
      printOut(vinciaName()+":init","Forcing Cf10 = Cf01 by averaging.");
      Cs10 = 0.5 * (Cf10 + Cf01);
      Cs01 = Cf10;
    }
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cp10 = Cp01 by averaging.");
      Cp10 = 0.5 * (Cp10 + Cp01);
      Cp01 = Cp10;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
		      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Mass corections
    denominatorMassPrefactor = settingsPtr->parm(vinciaName()
						 +":denominatorMassPrefactor");
    M1000 = settingsPtr->parm(vinciaName()+":M1000");
    M1010 = settingsPtr->parm(vinciaName()+":M1000");
    hasLinearM = (M1000 != 0.0 || M1010 != 0.0);
    M2000 = settingsPtr->parm(vinciaName()+":M2000");
    M2010 = settingsPtr->parm(vinciaName()+":M2010");
    hasSquareM = (M2000 != 0.0 || M2010 != 0.0);
    M1100 = settingsPtr->parm(vinciaName()+":M1100");
    M1110 = settingsPtr->parm(vinciaName()+":M1110");
    hasCrossM  = (M1100 != 0.0 || M1110 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // QQEmit::antennaFunction()

  double QQEmit::antennaFunction(double yij, double yjk, 
				    double mui, double muj, double muk,
				    int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;

    // No clue about massive gluons
    if (abs(muj) > TINY) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Ignoring gluon virtuality");
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > TINY || muk > TINY) 
      return antennaMassive(yij, yjk, mui, muk);    

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik  = 1./yij/yjk;
    double cij  = 1./yij;
    double cjk  = 1./yjk;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 2*(cij+cjk);
      if (doCollinear) term += yjk*cij + yij*cjk;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 2*cjk;
      if (doCollinear) term += yij*cjk; 
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 2*cij;
      if (doCollinear) term += yjk*cij;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;
    }
    
    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
    
  }

  //------------
  // QQEmit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QQEmit::antennaMassive(double yij, double yjk, 
				   double mui, double muk) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double mu2k = pow2(muk);
    double yik  = 1.0 - yij - yjk - mu2i - mu2k;

    // Compute eikonal, with mass correction
    double sum  = 2*yik/yij/yjk - 2*mu2i/pow2(yij) - 2*mu2k/pow2(yjk);
       
    // Denominator for collinear and finite terms
    double denom  = 1.0 - mu2i - mu2k + denominatorMassPrefactor*mui*muk;

    // Non-eikonal collinear terms
    if (useCollinearTermsSav) 
      sum += (yij/yjk + yjk/yij) / denom;
 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      if (hasFiniteTerms) {
	double C00 = 0.5 * ( Cs00 + Cf00 + 2*Cp00 );
	double C01 = 0.5 * ( Cs01 + Cf01 + 2*Cp01 );
	double C10 = 0.5 * ( Cs10 + Cf10 + 2*Cp10 );
	sum += (C00 + C10*yij + C01*yjk) / denom; 
      }

      // Add mass-dependent finite terms
      if (hasLinearM || hasSquareM || hasCrossM) {
	double term = 0.0;
	double ySum  = yij + yjk;
	if (hasLinearM) term += (mui + muk) * (M1000 + M1010*ySum);
	if (hasSquareM) term += (mu2i + mu2k) * (M2000 + M2010*ySum);
	if (hasCrossM) term  += mui*muk * (M1100 + M1110*ySum);
	sum += term / denom;
      }

    } // end of if (useFiniteTermsSav)
    
    // Return sum of antenna terms
    return sum;
    
  }

  //======================================================================

  //------------
  // QGEmit::init()

  bool QGEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = Antenna::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Set MIN variation specifically for this antenna
    // (MAX variation is fine just inheriting from base class)
    if (minMaxVar == -1) {
      Cs00 = -3;
      Cf00 = 3;
      Cp00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf10 = 0;
      Cf01 = 0;
      Cp10 = 0;
      Cp01 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
		      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Mass corections
    M100  = settingsPtr->parm(vinciaName()+":M100");
    M110  = settingsPtr->parm(vinciaName()+":M110");
    M101  = settingsPtr->parm(vinciaName()+":M101");
    hasLinearM = (M100 != 0.0 || M110 != 0.0 || M101 != 0.0);
    M200  = settingsPtr->parm(vinciaName()+":M200");
    M210  = settingsPtr->parm(vinciaName()+":M210");
    M201  = settingsPtr->parm(vinciaName()+":M201");
    M2m0  = settingsPtr->parm(vinciaName()+":M2m0");
    M2m1  = settingsPtr->parm(vinciaName()+":M2m1");
    hasSquareM = (M200 != 0.0 || M210 != 0.0 || M201 != 0.0 
		  || M2m0 != 0 || M2m1 != 0);
    T2mm2 = settingsPtr->parm(vinciaName()+":T2mm2");
    T3mm1 = settingsPtr->parm(vinciaName()+":T3mm1");
    T4mm1 = settingsPtr->parm(vinciaName()+":T4mm1");
    hasTildeM  = (T2mm2 != 0.0 || T3mm1 != 0.0 || T4mm1 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // QGEmit::antennaFunction()

  double QGEmit::antennaFunction(double yij, double yjk, 
				    double mui, double muj, double muk,
				    int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;

    // No clue about massive gluons
    if (muj != 0. || muk != 0.) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Ignoring gluon virtuality");
      muj = 0.0;
      muk = 0.0;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0.) 
      return antennaMassive(yij, yjk, mui);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // Hhow many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik      = 1./yij/yjk;
    double cij      = 1./yij;
    double cjk      = 1./yjk;
    double a        = alphaSav;
    double sectorK  = (isSectorSav) ? cjk/(1-yij) : 0.0;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite   = useFiniteTermsSav && hasFiniteTerms;
    bool doSector   = isSectorSav && useCollinearTermsSav && useSectorTermsSav;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doSector) term += sectorK;
      else if (doCollinear && a != 1.0) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 2*cij - 3*cjk;
      if (doCollinear) term += yjk*cij + 3*yij*cjk - pow2(yij)*cjk;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (pure sector) 
      if (doSector) {
	term = sectorK - cjk*(1. + yij + pow2(yij));
	if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
	if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 3*cjk;
      if (doCollinear) term += 3*yij*cjk - pow2(yij)*cjk; 
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 2*cij;
      if (doCollinear) term += yjk*cij;
      if (doSector)    term += sectorK;
      else if (doCollinear && a != 1.) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

      // +- > +-+ && -+ > -+- antennae (pure sector)
      if (doSector) {
	term = sectorK - cjk*(1. + yij + pow2(yij));
	if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
	if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
      }
    }
    
    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
    
  }

  //------------
  // QGEmit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QGEmit::antennaMassive(double yij, double yjk, double mui) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double yik  = 1.0 - yij - yjk - mu2i;

    // Compute eikonal, with mass correction
    double sum  = 2*yik/yij/yjk - 2*mu2i/pow2(yij);
       
    // Denominator for collinear and finite terms
    double denom  = 1.0 - mu2i;

    // Non-eikonal collinear terms
    if (useCollinearTermsSav) {
      sum += ( yjk/yij + yij/yjk * (1.-yij/denom) ) / denom; 
      // Massive sector terms (tentatively by replacing yij -> yij/denom)
      if (isSectorSav && useSectorTermsSav) 
	sum += (2./(1-yij/denom) - 2. + (1.-yij/denom)*yij/denom)/yjk;
    }

    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      if (hasFiniteTerms) {
	double C00 = 0.5 * ( Cs00 + Cf00 + 2*Cp00 );
	double C01 = 0.5 * ( Cs01 + Cf01 + 2*Cp01 );
	double C10 = 0.5 * ( Cs10 + Cf10 + 2*Cp10 );
	sum += (C00 + C10*yij + C01*yjk) / denom; 
      }

      // Add mass-dependent finite terms
      if (hasLinearM || hasSquareM || hasTildeM) {
	double term = 0.0;
	if (hasLinearM) term += mui * (M100 + M110*yij + M101*yjk);
	if (hasSquareM) term += mu2i * (M200 + M210*yij + M201*yjk
					+ M2m0/yij + M2m1*yjk/yij);
	if (hasTildeM)  
	  term += mu2i*yjk/yij/yik * (T2mm2*yjk + mui*T3mm1 + mu2i*T4mm1);

	sum += term / denom;
      }
      
    } // end of if (useFiniteTermsSav)
    
    // Return sum of antenna terms
    return sum;
    
  }

  //======================================================================

  //------------
  // GGEmit::init()

  bool GGEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = Antenna::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Set MIN variation for this specific antenna
    // (MAX variation is fine inheriting from base class)
    if (minMaxVar == -1) {
      Cs00 = -4;
      Cf00 = 3.7;
      Cp00 = 0.0;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
		      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Check whether finite terms are symmetric
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cs10 = Cs01 by averaging.");
      Cs10 = 0.5 * (Cs10 + Cs01);
      Cs01 = Cs10;
    }
    if (Cf10 != Cf01) {
      printOut(vinciaName()+":init","Forcing Cf10 = Cf01 by averaging.");
      Cs10 = 0.5 * (Cf10 + Cf01);
      Cs01 = Cf10;
    }
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cp10 = Cp01 by averaging.");
      Cp10 = 0.5 * (Cp10 + Cp01);
      Cp01 = Cp10;
    }

    // Return OK
    return isInit;
  }

  //------------
  // GGEmit::antennaFunction()
  
  double GGEmit::antennaFunction(double yij, double yjk, 
				    double mui, double muj, double muk,
				    int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // No clue about massive gluons
    if (mui != 0. || muj != 0. || muk != 0.) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Ignoring gluon virtuality.");
      mui = 0.0;
      muj = 0.0;
      muk = 0.0;
    }
    
    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // Hhow many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik      = 1./yij/yjk;
    double cij      = 1./yij;
    double cjk      = 1./yjk;
    double a        = alphaSav;
    double sectorI  = (isSectorSav) ? cij/(1-yjk) : 0.0;
    double sectorK  = (isSectorSav) ? cjk/(1-yij) : 0.0;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite   = useFiniteTermsSav && hasFiniteTerms;
    bool doSector   = isSectorSav && useCollinearTermsSav && useSectorTermsSav;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doSector) term += sectorI + sectorK;
      else if (doCollinear && a != 1.0) 
	term += (1.-a)*(cjk + cij) - 2*(1.-a)*(yij*cjk + yjk*cij);
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 3*cij - 3*cjk;
      if (doCollinear) 
	term += 3*yjk*cij + 3*yij*cjk - pow2(yij)*cjk - pow2(yjk)*cij;
      if (doFinite) term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (pure sector) 
      if (doSector) {
	term = sectorK - cjk*(1. + yij + pow2(yij));
	if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
	if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

      // ++ > -++ && -- > +-- antennae (pure sector) 
      if (doSector) {
	term = sectorI - cij*(1. + yjk + pow2(yjk));
	if ( RH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;
	if ( LH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      }

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 3*cjk;
      if (doCollinear) term += 3*yij*cjk - pow2(yij)*cjk; 
      if (doSector)    term += sectorI;
      else if (doCollinear && a != 1.) term += (1.-a)*cij - 2*(1.-a)*yjk*cij;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 3*cij;
      if (doCollinear) term += 3*yjk*cij - pow2(yjk)*cij;
      if (doSector)    term += sectorK;
      else if (doCollinear && a != 1.) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

      // +- > +-+ && -+ > -+- antennae (pure sector)
      if (doSector) {
	term = sectorK - cjk*(1. + yij + pow2(yij));
	if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
	if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
      }

      // +- > -+- && -+ > +-+ antennae (pure sector)
      if (doSector) {
	term = sectorI - cij*(1. + yjk + pow2(yjk));
	if ( RH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
	if ( LH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

    }
    
    // Return helicity sum, averaged over initial helicities 
    return hSum/nAvg;
    
  }

  //======================================================================

  //------------
  // QGSplit::init()

  bool QGSplit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = Antenna::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // MIN variation (zero finite terms)
    if (minMaxVar == -1) {
      Cs00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf00 = 0;
      Cf10 = 0;
      Cf01 = 0;      
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0);
    
    // Mass corections
    M1000 = settingsPtr->parm(vinciaName()+":M1000");
    M1010 = settingsPtr->parm(vinciaName()+":M1010");
    M1001 = settingsPtr->parm(vinciaName()+":M1001");
    hasLinearM = (M1000 != 0.0 || M1010 != 0.0 || M1001 != 0.0);
    M2000 = settingsPtr->parm(vinciaName()+":M2000");
    M2010 = settingsPtr->parm(vinciaName()+":M2010");
    M2001 = settingsPtr->parm(vinciaName()+":M2001");
    hasSquareM = (M2000 != 0.0 || M2010 != 0.0 || M2001 != 0.0);
    T00   = settingsPtr->parm(vinciaName()+":T00");
    T10   = settingsPtr->parm(vinciaName()+":T10");
    T01   = settingsPtr->parm(vinciaName()+":T01");
    hasSquareD = (T00 != 0.0 || T10 != 0.0 || T01 != 0.0);
    T1000 = settingsPtr->parm(vinciaName()+":T1000");
    T1010 = settingsPtr->parm(vinciaName()+":T1010");
    T1001 = settingsPtr->parm(vinciaName()+":T1001");
    T2000 = settingsPtr->parm(vinciaName()+":T2000");
    T2010 = settingsPtr->parm(vinciaName()+":T2010");
    T2001 = settingsPtr->parm(vinciaName()+":T2001");
    hasTildeM  = (T1000 != 0.0 || T1010 != 0.0 || T1001 != 0.0
		  || T2000 != 0.0 || T2010 != 0.0 || T2001 != 0.0);
    T0200 = settingsPtr->parm(vinciaName()+":T0200");
    T0210 = settingsPtr->parm(vinciaName()+":T0210");
    T0201 = settingsPtr->parm(vinciaName()+":T0201");	
    hasQuarticD = (T0200 != 0.0 || T0210 != 0.0 || T0201 != 0.0); 
    
    // Return OK
    return isInit;
  }

  //------------
  // QGSplit::antennaFunction()
  double QGSplit::antennaFunction(double yij, double yjk, 
				  double mui, double muj, double muk,
				  int hA, int hB, int hi, int hj, int hk) {

    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // Require equal jk masses
    if ( abs(muj-muk) > TINY ) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Warning: forcing mq = mqbar.");
      cout<<"QGSPLIT: y(ij,jk) = "<<yij<<" "<<yjk<<" mu(i,j,k) = "
	  <<scientific<<mui<<" "<<muj<<" "<<muk<<endl;
      muk = muj;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0. || muj > 0.) 
      return antennaMassive(yij, yjk, mui, muj);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double cjk  = 1./yjk;
    double tf   = cjk * 0.5 * pow2(yij);
    double ts   = cjk * ( 0.5 - yij ) + tf;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +-+ && -- > -+- antennae (s)
      double term = ts;
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae (s)
      double term = ts; 
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-+ && -+ > -+- antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

    }
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) hSum *= 2.;    

    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
  }

  //------------
  // QGSplit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QGSplit::antennaMassive(double yij, double yjk, 
				    double mui, double muj) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double mu2j = pow2(muj);
    double yik  = 1.0 - yij - yjk - mu2i - 2*mu2j;

    // Gluon propagator and sIK^2 denominator factors
    double prop  = yjk + 2*mu2j;
    double denom = pow2(1. - mu2i);

    // Compute eikonal, with mass correction
    double sum  = 0.5 / prop * ( (pow2(yik) + pow2(yij))/denom
				 + 2*mu2j/prop );
				 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      // (Here written with explicit averaging and summing to 
      // avoid confusion)
      if (hasFiniteTerms) {
	double C00 = 0.5 * ( 2*Cs00 + 2*Cf00);
	double C01 = 0.5 * ( 2*Cs01 + 2*Cf01 );
	double C10 = 0.5 * ( 2*Cs10 + 2*Cf10 );
	sum += yjk / prop * (C00 + C10*yij + C01*yjk) / denom; 
      }

      // Add daughter-mass-dependent finite terms
      if (hasSquareD || hasQuarticD) {
	double term = 0.0;
	if (hasSquareD) term += mu2j * (T00 + T10*yij + T01*yjk);
	if (hasQuarticD) term += pow2(mu2j) * (T0200 + T0210*yij + T0201*yjk);
	sum += term / prop / denom;
      }

      // Add parent-mass-dependent finite terms
      if (hasLinearM || hasSquareM || hasTildeM) {
	double term = 0.0;
	if (hasLinearM) term += mui * (M1000 + M1010*yij + M1001*yjk);
	if (hasSquareM) term += mu2i * (M2000 + M2010*yij + M2001*yjk);
	if (hasTildeM)  
	  term += mu2j*mui*(T1000 + M1010*yij + M1001*yjk 
			    + mui*(T2000 + T2010*yij + T2001*yjk));
	sum += term / prop / denom;
      }
      
    } // end of if (useFiniteTermsSav)
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) sum *= 2;
    
    // Return sum of antenna terms
    return sum;
    
  }

//======================================================================

  //------------
  // GGSplit::init()

  bool GGSplit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = Antenna::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // MIN variation (zero finite terms)
    if (minMaxVar == -1) {
      Cs00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf00 = 0;
      Cf10 = 0;
      Cf01 = 0;      
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
		      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0);
    
    // Mass corections
    T00   = settingsPtr->parm(vinciaName()+":T00");
    T10   = settingsPtr->parm(vinciaName()+":T10");
    T01   = settingsPtr->parm(vinciaName()+":T01");
    hasSquareD = (T00 != 0.0 || T10 != 0.0 || T01 != 0.0);
    M200 = settingsPtr->parm(vinciaName()+":M200");
    M210 = settingsPtr->parm(vinciaName()+":M210");
    M201 = settingsPtr->parm(vinciaName()+":M201");
    hasQuarticD = (M200 != 0.0 || M210 != 0.0 || M201 != 0.0); 
    
    // Return OK
    return isInit;
  }

  //------------
  // GGSplit::antennaFunction()

  double GGSplit::antennaFunction(double yij, double yjk, 
				    double mui, double muj, double muk,
				    int hA, int hB, int hi, int hj, int hk) {

    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // Ignore gluon mass
    if (mui != 0.0) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Ignoring gluon virtuality");
      mui = 0.0;
    }

    // Require equal jk masses (if no k given, set k = j)
    if ( abs(muj-muk) > TINY ) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
				 "Warning: forcing mq = mqbar.");
      muk = muj;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0. || muj > 0.) 
      return antennaMassive(yij, yjk, muj);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double cjk  = 1./yjk;
    double tf   = cjk * 0.5 * pow2(yij);
    double ts   = cjk * ( 0.5 - yij ) + tf;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +-+ && -- > -+- antennae (s)
      double term = ts;
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae (s)
      double term = ts; 
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-+ && -+ > -+- antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
    }
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) hSum *= 2.;

    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
  }


  //------------
  // GGSplit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double GGSplit::antennaMassive(double yij, double yjk, double muj) {

   // Compute third invariant
    double mu2j = pow2(muj);
    double yik  = 1.0 - yij - yjk - 2*mu2j;

    // Gluon propagator factor
    double prop  = yjk + 2*mu2j;

    // Compute eikonal, with mass correction
    double sum  = 0.5 / prop * ( (pow2(yik) + pow2(yij)) + 2*mu2j/prop );
				 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      // (Here written with explicit averaging and summing to 
      // avoid confusion)
      if (hasFiniteTerms) {
	double C00 = 0.5 * ( 2*Cs00 + 2*Cf00);
	double C01 = 0.5 * ( 2*Cs01 + 2*Cf01 );
	double C10 = 0.5 * ( 2*Cs10 + 2*Cf10 );
	sum += yjk / prop * (C00 + C10*yij + C01*yjk); 
      }

      // Add daughter-mass-dependent finite terms
      if (hasSquareD || hasQuarticD) {
	double term = 0.0;
	if (hasSquareD) term += mu2j * (T00 + T10*yij + T01*yjk);
	if (hasQuarticD) term += pow2(mu2j) * (M200 + M210*yij + M201*yjk);
	sum += term / prop;
      }

    } // end of if (useFiniteTermsSav)
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) sum *= 2;
    
    // Return sum of antenna terms
    return sum;
    
  } 

  // End inherited Antenna classes
  //======================================================================


} // End Vincia Namespace
  
